#
#  peppercornenumerator/enumerator.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division
from builtins import input

import logging
log = logging.getLogger(__name__)

import sys
import math

from peppercornenumerator.condense import PepperCondensation
from peppercornenumerator.objects import PepperMacrostate, PepperComplex
from peppercornenumerator.objects import DSDDuplicationError, DSDObjectsError
from peppercornenumerator.utils import PeppercornUsageError
from peppercornenumerator.input import read_pil, read_seesaw
from peppercornenumerator.output import write_pil, write_sbml
import peppercornenumerator.reactions as reactlib


# There should be better control of this limit -- only set it when
# necessary (Erik Winfree based on Chris Thachuk's advice...)
sys.setrecursionlimit(20000)


FAST_REACTIONS = {1: [reactlib.bind11, reactlib.open,
                      reactlib.branch_3way, reactlib.branch_4way]}
"""
Dictionary of reaction functions considered *fast* for a given "arity".
Keys are arities (e.g. 1 = unimolecular, 2 = bimolecular, 3 = trimolecular,
etc.), and values are lists of reaction functions. Currently, only
unimolecular fast reactions (arity = 1) are supported.
"""

SLOW_REACTIONS = {
    1: [],
    2: [reactlib.bind21]
}
"""
A dictionary of reaction functions considered *slow* for a given "arity".
Keys are arities (e.g. 1 = unimolecular, 2 = bimolecular, 3 = trimolecular,
etc.), and values are lists of reaction functions. Currently, only
bimolecular reactions (arity = 2) are by default slow. However, when using
k-fast, also unimolecular reactions can fall into the *slow* regime.
"""

class PolymerizationError(Exception):
    """Error class to catch polymerization."""

    def __init__(self, msg, val=None):
        self.message = msg
        if val :
            self.message += " ({})".format(val)
        super(PolymerizationError, self).__init__(self.message) 

class Enumerator(object):
    """
    The main enumerator object, collecting all the parameters to enumerate the
    Chemical Reaction Network (CRN) from an initial set of domain-level
    complexes.  Enumerators have immutable starting conditions. 

    Args:
        initial_complexes(list[PepperComplex()]): The list of complexes
            presenting a starting condition for the enumeration process.
        initial_reactions(list[PepperReaction()], optional): An optional list
            of Reactions to be added to the enumeration results. These 
            reactions will be considered when finding SCCs and when condensing
            the CRN after enumeration.
    """

    def __init__(self, initial_complexes, initial_reactions=None):
        # System initialization
        self._initial_complexes = initial_complexes
        assert isinstance(initial_complexes, list)
        for cplx in initial_complexes:
            if not cplx.is_connected:
                raise PeppercornUsageError('Initial complex is not connected: {}'.format(cplx))

        # list containing all detailed reactions after enumeration
        if initial_reactions:
            self._reactions = set(initial_reactions)
        else :
            self._reactions = set()

        # A list of "known" complexes, 
        #   e.g. expected to show up during enumeration.
        # This is useful when chosing representatives of resting macrostates.
        self._named_complexes = list(PepperComplex.MEMORY.values())
        
        # Enumeration results
        self._complexes = None
        self._resting_complexes = None
        self._transient_complexes = None
        self._resting_macrostates = None

        # Condensation
        self.condensation = None
        self._detailed_reactions = None
        self._condensed_reactions = None

        # Polymerization settings to prevent infinite looping
        self._max_complex_count = max(200, len(self._initial_complexes))
        self._max_reaction_count = max(1000, len(self._reactions))

        # Settings for reaction enumeration. 
        self.max_helix = True
        self.reject_remote = False
        self._release_11 = 7
        self._release_12 = 7
        self._max_complex_size = 6

        # Rate-dependent enumeration settings:
        #
        # Set separation of timescales for *unimolecular* reactions.
        #
        #  ignore-reaction | slow-reaction | fast-reaction
        # -----------------|---------------|---------------> rate
        #                 k_slow          k_fast
        #
        # Default: OFF = all unimolecular reactions are fast or negligible
        self._k_slow = 0 
        self._k_fast = 0

        # Rate parameter adjustments
        self.dG_bp = -1.7 # adjust bimolecular binding strength

        # Debugging features
        self.DFS = True
        self.interactive = False
        self.interruptible = False

        # Not officially supported parameters
        self.local_elevation = False
        self.p_min = 0 # minimum complex steady state probability to engage in slow reaction.

    @property
    def max_complex_size(self):
        return self._max_complex_size

    @max_complex_size.setter
    def max_complex_size(self, value):
        assert isinstance(value, int)
        if not all(cx.size <= value for cx in self._initial_complexes):
            msize = max([c.size for c in self._initial_complexes])
            raise PeppercornUsageError(
                    'Maximum complex size must include all input complexes.',
                    'Set to at least: {}'.format(msize))
        self._max_complex_size = value

    @property
    def max_reaction_count(self):
        return self._max_reaction_count

    @max_reaction_count.setter
    def max_reaction_count(self, value):
        assert isinstance(value, int) and value >= len(self._reactions)
        self._max_reaction_count = value

    @property
    def max_complex_count(self):
        return self._max_complex_count

    @max_complex_count.setter
    def max_complex_count(self, value):
        assert isinstance(value, int) and value >= len(self._initial_complexes)
        self._max_complex_count = value

    @property
    def k_slow(self):
        return self._k_slow

    @k_slow.setter
    def k_slow(self, value):
        if value > 0:
            (rc, k_rc) = (0, None)
            while True:
                rc += 1
                k_rc = reactlib.opening_rate(rc, dG_bp = self.dG_bp)
                if k_rc < value: break
            self.release_cutoff = max(rc, self.release_cutoff)

        if 0 < self.k_fast < value :
            raise PeppercornUsageError('k-slow must not be bigger than k-fast.')

        self._k_slow = value

    @property
    def k_fast(self):
        return self._k_fast

    @k_fast.setter
    def k_fast(self, value):
        if 0 < value < self.k_slow:
            raise PeppercornUsageError('k-fast must not be smaller than k-slow.')
        self._k_fast = value

    @property
    def release_cutoff(self):
        if self._release_11 != self._release_12 :
            raise PeppercornUsageError('Ambiguous release cutoff request.')
        return self._release_11

    @release_cutoff.setter
    def release_cutoff(self, value):
        assert isinstance(value, int) and value >= 0
        self._release_11 = value
        self._release_12 = value

    @property
    def release_cutoff_1_1(self):
        return self._release_11

    @release_cutoff_1_1.setter
    def release_cutoff_1_1(self, value):
        assert isinstance(value, int) and value >= 0
        self._release_11 = value

    @property
    def release_cutoff_1_2(self):
        return self._release_12

    @release_cutoff_1_2.setter
    def release_cutoff_1_2(self, value):
        assert isinstance(value, int) and value >= 0
        self._release_12 = value

    @property
    def remote_migration(self):
        print(DeprecationWarning('Peppercorn>=v0.7: use Enumerator.reject_remote instead of Enumerator.remote_migration.'))
        return not self.reject_remote

    @remote_migration.setter
    def remote_migration(self, remote):
        assert isinstance(remote, bool)
        print(DeprecationWarning('Peppercorn>=v0.7: use Enumerator.reject_remote instead of Enumerator.remote_migration.'))
        self.reject_remote = not remote

    @property
    def max_helix_migration(self):
        print(DeprecationWarning('Peppercorn>=v0.7: use Enumerator.max_helix instead of Enumerator.max_helix_migration.'))
        return self.max_helix

    @max_helix_migration.setter
    def max_helix_migration(self, max_helix):
        print(DeprecationWarning('Peppercorn>=v0.7: use Enumerator.max_helix instead of Enumerator.max_helix_migration.'))
        self.max_helix = max_helix

    @property
    def initial_complexes(self):
        """
        Complexes present in the system's initial configuration
        """
        return self._initial_complexes[:]

    @property
    def domains(self):
        domains = set()
        for cplx in self._initial_complexes:
            [domains.add(d) for d in cplx.domains]
        return list(domains)

    @property
    def reactions(self):
        """
        List of reactions enumerated. :py:meth:`.enumerate` must be
        called before access.
        """
        return list(self._reactions)

    @property
    def detailed_reactions(self):
        return list(self._reactions)

    @property
    def condensed_reactions(self):
        if self.condensation is None:
            self.condense()
        return self.condensation.condensed_reactions

    @property
    def resting_sets(self):
        print("# Deprecated function: Enumerator.resting_sets. Replace with Enumerator.resting_macrostates")
        return self.resting_macrostates

    @property
    def resting_macrostates(self):
        """
        List of resting states enumerated. :py:meth:`.enumerate` must be
        called before access.
        """
        if self._resting_macrostates is None:
            raise PeppercornUsageError("enumerate not yet called!")
        return self._resting_macrostates[:]

    @property
    def complexes(self):
        """
        List of complexes enumerated. :py:meth:`.enumerate` must be
        called before access.
        """
        if self._complexes is None:
            raise PeppercornUsageError("enumerate not yet called!")
        return self._complexes[:]

    @property
    def resting_complexes(self):
        """
        List of complexes enumerated that are within resting states.
        :py:meth:`.enumerate` must be called before access.
        """
        if self._resting_complexes is None:
            raise PeppercornUsageError("enumerate not yet called!")
        return self._resting_complexes[:]

    @property
    def transient_complexes(self):
        """
        List of complexes enumerated that are not within resting sets (e.g.
        complexes which are transient). :py:meth:`.enumerate` must be
        called before access.
        """
        if self._transient_complexes is None:
            raise PeppercornUsageError("enumerate not yet called!")
        return self._transient_complexes[:]

    def __eq__(self, other):
        raise NotImplementedError('No notion of equality implemented for Enumerator objects.')

    def __ne__(self, other):
        return not (self == other)

    def dry_run(self):
        """
        Make it look like you've enumerated, but actually do nothing...
        """
        if self._complexes:
            raise PeppercornUsageError('Cannot call dry-run after enumeration!')

        rep = set(self._named_complexes) if self._named_complexes else set(self._initial_complexes)
        rxs = [r for r in list(self._reactions) if len(r.reactants) == 1]
        info = segment_neighborhood(self.initial_complexes, rxs, self.p_min, represent=rep)

        self._complexes = self.initial_complexes
        self._resting_complexes = info['resting_complexes']
        self._transient_complexes = info['transient_complexes']
        self._resting_macrostates = info['resting_macrostates']

    def enumerate(self):
        """
        Generates the reaction graph consisting of all complexes reachable from
        the initial set of complexes. Produces a full list of :py:meth:`complexes`, resting
        sets, and :py:meth:`reactions, which are stored in the associated members of this
        class.
        """

        # Will be called once enumeration halts, either because it's finished or
        # because too many complexes/reactions have been enumerated
        def finish(premature=False):
            # copy E and T into complexes
            self._complexes += (self._E)
            self._complexes += (self._T)

            # preserve resting and transient complexes separately
            self._transient_complexes = self._T
            self._resting_complexes = self._E

            # If we're bailing because of too many reactions or complexes, search
            # self._reactions to ensure there are no reactions which contain
            # products that never made it into self._complexes...
            if premature:
                self._resting_complexes += self._S
                self._complexes += self._S
                complexes = set(self._complexes)

                rm_reactions = []
                for reaction in self.reactions:
                    #NOTE: This should only matter in the Ctrl-C case, right?
                    reaction_ok = all((prod in complexes) for prod in reaction.products) and \
                                  all((reac in complexes) for reac in reaction.reactants)

                    if reaction_ok:
                        pass
                    else :
                        rm_reactions.append(reaction)

                self._reactions -= set(rm_reactions)

        # List E contains enumerated resting complexes. Every time a new
        # complex is added (from S), all cross reactions with other resting
        # complexes are enumerated. Complexes remain in this list throughout
        # function execution, the products of slow reactions go into list B.
        self._E = []

        # List S contains newly determined resting complexes after all fast
        # reactions have been enumerated. They will be moved to E and thereby
        # tested for cross reactions with set E. 
        self._S = []

        # List T contains newly determined transient states after fast-reaction
        # enumeration. These complexes will remain in this list throughout
        # function execution.
        self._T = []

        # List N contains the neighborhood of some initial complexes with all
        # fast reactions enumerated. Complexes in N have not yet been
        # characterized as transient or resting sets.
        self._N = []

        # List F contains components of the current 'neighborhood' which have
        # not yet been reactants for potential fast reactions.  They will be
        # moved to N once they were enumerated.
        self._F = []

        # List B contains initial complexes, or products of bimolecular
        # reactions that have had no reactions enumerated yet. They will be
        # moved to F for their fast neighborhood to be enumerated.
        self._B = self.initial_complexes

        self._complexes = []
        self._resting_macrostates = []

        def do_enumerate():
            log.debug("Fast reactions from initial complexes ...")
            while len(self._B) > 0:
                # Generate a neighborhood from `source`
                source = self._B.pop()
                self.process_neighborhood(source)

            # Consider slow reactions between resting set complexes
            log.debug("Slow reactions between resting set complexes ...")
            while len(self._S) > 0:

                # Find slow reactions from 'element'
                if self.DFS:
                    element = self._S.pop()
                else:
                    element = self._S.pop(0)

                log.debug("Slow reactions from complex {:s} ({:d} remaining in S)".format(
                    str(element), len(self._S)))
                slow_reactions = self.get_slow_reactions(element)
                self._E.append(element)

                # Find the new complexes which were generated
                self._B = self.get_new_products(slow_reactions)
                self._reactions |= set(slow_reactions)
                log.debug("Generated {:d} new slow reactions".format(len(slow_reactions)))
                log.debug("Generated {:d} new products".format(len(self._B)))

                # Display new reactions in interactive mode
                if self.interactive:
                    self.reactions_interactive(element, slow_reactions, 'slow')

                # Now find all complexes reachable by fast reactions from these new complexes
                while len(self._B) > 0:

                    # Check whether too many complexes have been generated
                    if (len(self._E) + len(self._T) + len(self._S) > self._max_complex_count):
                        raise PolymerizationError("Too many complexes enumerated!", 
                                len(self._E) + len(self._T) + len(self._S))

                    # Check whether too many reactions have been generated
                    if (len(self._reactions) > self._max_reaction_count):
                        raise PolymerizationError("Too many reactions enumerated!", 
                                len(self._reactions))

                    # Generate a neighborhood from `source`
                    source = self._B.pop()
                    self.process_neighborhood(source)

        if self.interruptible:
            try:
                do_enumerate()
                finish()
            except KeyboardInterrupt:
                log.warning("Interrupted; gracefully exiting...")
                finish(premature=True)
            except PolymerizationError as err:
                log.exception(err)
                log.error("Polymerization error; gracefully exiting...")
                finish(premature=True)
        else:
            do_enumerate()
            finish()

    def condense(self):
        self.condensation = PepperCondensation(self)
        self.condensation.condense()

    def to_pil(self, filename = None, **kwargs):
        if filename:
            with open(filename, 'w') as pil:
                write_pil(self, fh=pil, **kwargs)
            return ''
        else:
            return write_pil(self, **kwargs)

    def to_sbml(self, filename = None, **kwargs):
        if filename:
            with open(filename, 'w') as sbml:
                write_sbml(self, fh = sbml, **kwargs)
            return ''
        else:
            return write_sbml(self, **kwargs)

    def reactions_interactive(self, root, reactions, rtype='fast'):
        """
        Prints the passed reactions as a kernel string, then waits for keyboard
        input before continuing.
        """
        print("{} = {} ({})".format(root.name, root.kernel_string, rtype))
        if len(reactions) is 0:
            print("(No {} reactions)".format(rtype))
        for r in reactions:
            print("{} ({}: {})".format(r, r.rtype, r.const))
            print(" {}".format(r.kernel_string))
        input("\n[Press Enter to continue...]")

    def process_neighborhood(self, source):
        """ Enumerate neighborhood of fast reactions.

        Takes a single complex, generates the 'neighborhood' of complexes
        reachable from that complex through fast reactions, classifies these
        complexes as transient or resting complexes, and modifies the global
        lists and list of reactions accordingly.

        Args:
            source (:obj:`PepperComplex`): Initial complex to generate a
                neighborhood of fast reactions.
        """

        assert len(self._F) == 0
        self._F = [source]

        # N_reactions holds reactions which are part of the current neighborhood
        N_reactions = []

        log.debug("Processing neighborhood: %s" % source)

        interrupted = False
        try:
            while len(self._F) > 0 :

                # Find fast reactions from `element`
                element = self._F.pop()
                log.debug("Fast reactions from {:s}... ({:d} remaining in F)".format(
                    str(element), len(self._F)))

                # Return valid fast reactions:
                reactions = self.get_fast_reactions(element)
                   
                # Add new products to F
                new_products = self.get_new_products(reactions)
                self._F += (new_products)

                # Add new reactions to N_reactions
                N_reactions += (reactions)
                self._N.append(element)

                log.debug("Generated {:d} new fast reactions.".format(len(reactions)))
                log.debug("Generated {:d} new products.".format(len(new_products)))
                   
                # Display new reactions in interactive mode
                if self.interactive:
                    self.reactions_interactive(element, reactions, 'fast')

        except KeyboardInterrupt:
            log.warning("Exiting neighborhood %s prematurely..." % source)
            if self.interruptible:
                interrupted = True
            else :
                raise KeyboardInterrupt

        log.debug("In neighborhood %s..." % source)
        log.debug("Segmenting %d complexes and %d reactions" %
                      (len(self._N), len(N_reactions)))

        # Now segment the neighborhood into transient and resting complexes
        # by finding the strongly connected components.
        rep = set(self._named_complexes) if self._named_complexes else set(self._initial_complexes)
        segmented_neighborhood = segment_neighborhood(self._N, N_reactions, 
                                                      p_min = self.p_min,
                                                      represent = rep)

        # Resting complexes are added to S
        self._S += (segmented_neighborhood['resting_complexes'])

        # Transient complexes are added to T
        self._T += (segmented_neighborhood['transient_complexes'])

        # Resting macrostates are added to global list
        self._resting_macrostates += (segmented_neighborhood['resting_macrostates'])

        # Reactions from this neighborhood are added to the list
        self._reactions |= set(N_reactions)

        # Reset neighborhood
        log.debug("Generated {:d} new fast reactions".format(len(N_reactions)))
        log.debug("Generated {:d} new complexes: ({:d} transient, {:d} resting)".format(
            len(self._N), len(segmented_neighborhood['transient_complexes']), 
                len(segmented_neighborhood['resting_complexes'])))
        self._N = []

        log.debug("Generated {:d} resting macrostates".format(
            len(segmented_neighborhood['resting_macrostates'])))
        log.debug("Done processing neighborhood: {:s}".format(str(source)))

        if interrupted:
            raise KeyboardInterrupt

    def get_slow_reactions(self, complex):
        """
        Returns a list of slow reactions possible using complex and other
        complexes in list E as reagents.

        This only supports unimolecular and bimolecular reactions. Could be
        extended to support arbitrary reactions.
        """

        maxsize = self._max_complex_size

        reactions = []

        # Do unimolecular reactions that are always slow... not supported
        # anymore...
        assert SLOW_REACTIONS[1] == []

        # Do unimolecular reactions that are sometimes slow
        if self._k_fast > self._k_slow:
            for move in FAST_REACTIONS[1]:
                if move.__name__ == 'open':
                    move_reactions = move(complex, 
                            max_helix = self.max_helix, 
                            release_11 = self._release_11,
                            release_1N = self._release_12,
                            dG_bp = self.dG_bp)
                else :
                    move_reactions = move(complex, 
                            max_helix = self.max_helix, 
                            remote = not self.reject_remote)

                if self.local_elevation:
                    reactions += [rxn for rxn in move_reactions if self._k_slow <= local_elevation_rate(rxn) < \
                                self._k_fast]
                else :
                    reactions += [r for r in move_reactions if self._k_slow <= r.const < self._k_fast]
            for rxn in reactions:
                log.info('adding unimolecular slow reaction {}'.format(rxn.full_string()))

        # Do bimolecular reactions
        for move in SLOW_REACTIONS[2]:
            reactions += (move(complex, complex))
            for complex2 in self._E:
                reactions += (move(complex, complex2))

        # And now remove all over max-complex-size
        valid_reactions = []
        for rxn in reactions: 
            if maxsize and not all(p.size <= maxsize for p in rxn.products) :
                log.warning("Product complex size (={}) larger than --max-complex-size(={}). Ignoring slow reaction {}!".format(max([p.size for p in rxn.products]), maxsize, str(rxn)))
                continue
            valid_reactions.append(rxn)

        return valid_reactions

    def get_fast_reactions(self, cplx, rtypes = None, restrict=True):
        """Returns a list of fast reactions possible using complex as a reagent.

        Args:
            cplx: A reactant complex.
            rtypes (str, optional): Only reactions of a certain type.
            restrict (bool, optional): Use golbal parameters release_cutoff, 


        """

        maxsize = self._max_complex_size

        # Do unimolecular reactions
        reactions = []
        for move in FAST_REACTIONS[1]:
            if rtypes and move.__name__ not in rtypes:
                continue
            if move.__name__ == 'open':
                if restrict:
                    move_reactions = move(cplx, 
                        max_helix = self.max_helix, 
                        release_11 = self._release_11,
                        release_1N = self._release_12,
                        dG_bp = self.dG_bp)
                else :
                    move_reactions = move(cplx, 
                        max_helix = self.max_helix, 
                        release_11 = 0, release_1N = 0,
                        dG_bp = self.dG_bp)
            else :
                move_reactions = move(cplx, 
                        max_helix = self.max_helix, 
                        remote = not self.reject_remote)
            
            for rxn in move_reactions: 
                if maxsize and any(p.size > maxsize for p in rxn.products):
                    log.warning("Product complex size (={}) larger than --max-complex-size(={}). Ignoring fast reaction {}!".format( max([p.size for p in rxn.products]), maxsize, str(rxn)))
                    continue
                reactions.append(rxn)

        if restrict: # apply the k-fast / k-slow filter
            if self.local_elevation:
                # let's first remove reactions that cannot pass...
                reactions = [rxn for rxn in reactions if rxn.const >= self._k_slow]
                assert cplx._elevation is None
                el = self.get_local_elevation(cplx, reactions)
                cplx._elevation = el
                reactions = [rxn for rxn in reactions if local_elevation_rate(rxn, el) >= \
                                max(self._k_slow, self._k_fast)]
            else :
                reactions = [rxn for rxn in reactions if rxn.const >= max(self._k_slow, self._k_fast)]

        return reactions

    def get_new_products(self, reactions):
        """
        Checks the products in the list of reactions. Returns the new complexes in a list.

        """
        new_products = set()

        ESTNF = set(self._E + self._S + self._T + self._N + self._F)
        B = set(self._B)

        # Loop over every reaction
        for reaction in reactions:

            # Check every product of the reaction to see if it is new
            for (i, product) in enumerate(reaction.products):

                # This should have been checked earlier...
                assert product.size <= self._max_complex_size

                if product in B:
                    # If the product is in list B, then we need to remove it from
                    # that list so that it can be enumerated for self-interactions
                    # as part of this neighborhood
                    self._B.remove(product)
                    B.remove(product)

                # has not been enumerated...
                if product not in ESTNF:
                   new_products.add(product)

        assert (ESTNF - new_products) == ESTNF
        assert (ESTNF - B) == ESTNF

        return list(new_products)

    def get_local_elevation(self, cplx, reactions):
        """Calculate the local elevation of a complex.

        Local elevation is the inverse of the maximum free energy gain when
        applying a sequence of compatible moves. In other words, the energy
        difference between this complex and its gradient decent local minimum.
        Because our rate model is insufficient, this is just an approximation: 

            (1) Take all unary open/bind reactions, and their reverse reaction. 
            (2) Calculate the free energy change for each reversible reaction: 
                dG = ln(k1/k2)
            (3) Solve the max-clique problem to identify compatible downhill
                reactions. That means, reactions that can be applied
                successively leading to the same local minimum. 
            (4) Calculate the cummulative free energy change for every path
                that leads to a local minimum. The maximum value is the local
                elevation.

        """

        def rev_rtype(rtype, arity):
            """Returns the reaction type of a corresponding reverse reaction. """
            if rtype == 'open' and arity == (1,1):
                return 'bind11'
            elif rtype == 'open' and arity == (1,2):
                return 'bind21'
            elif rtype == 'bind11' or rtype == 'bind21':
                return 'open'
            else:
                raise NotImplementedError

        def elevation(rxn):
            assert rxn.arity == (1,1)
            if rxn.reverse_reaction is None:
                rr = rev_rtype(rxn.rtype, rxn.arity)
                r = self.get_fast_reactions(rxn.products[0], rtypes = [rr], restrict=False)
                r = [x for x in r if sorted(x.products) == sorted(rxn.reactants)]
                assert len(r) == 1
                if len(r) == 1:
                    rxn.reverse_reaction = r[0]
                    r[0].reverse_reaction = rxn
                else :
                    rxn.reverse_reaction = False
            
            dG = math.log(rxn.const/rxn.reverse_reaction.const)
            # downhill reaction has dG > 0
            return dG if dG > 0 else 0

        def try_move(reactant, rotations, rtype, meta):
            """
            Args:
                rectant: Is the product of a previous rection
                rotations: Rotations needed to rotate this product back in reactant form
                rtype: the type of the new reaction
                meta: the find_on_loop parameters for the new reaction
            """
            
            (invader, target, x_linker, y_linker) = meta

            # First, make sure invader and target locations map to the reactant-rotation.
            if rotations is not None and rotations != 0:
                invaders = [reactant.rotate_location(x, rotations) for x in invader.locs]
                targets = [reactant.rotate_location(x, rotations) for x in target.locs]
            else :
                invaders = list(invader.locs)
                targets = list(target.locs)

            # forbid max helix.... :-(
            assert len(invaders) == 1
            assert len(targets) == 1

            def triple(loc):
                return (reactant.get_domain(loc), reactant.get_structure(loc), loc)

            if rtype == 'bind11':
                # fore i in invaders...
                results = reactlib.find_on_loop(reactant, invaders[0], reactlib.filter_bind11)
            elif rtype == 'open':
                return reactant.get_structure(invaders[0]) == targets[0]
            else : 
                raise NotImplementedError('rtype not considered in local neighborhood: {}'.format(
                    rtype))

            targets = list(map(triple, targets))
            for [s,x,t,y] in results:
                if t == targets:
                    return True
            return False
        
        # Calculate the local elevation of a complex using only 1-1 reactions.
        compat = dict()
        for rxn in reactions:
            # exclude open and branch-migration with arity 1,2
            if rxn.arity != (1,1): continue
            if rxn.rtype in ('branch-3way', 'branch-4way'): continue
            compat[rxn] = set()
            for other in reactions:
                if rxn == other: continue
                if other.arity != (1,1): continue
                if other.rtype in ('branch-3way', 'branch-4way'): continue
                # Try to append the other reaction to this reaction
                success = try_move(rxn.products[0], rxn.rotations, other.rtype, other.meta)
                if success:
                    # We can apply other to the product of rxn
                    compat[rxn].add(other)

        # https://en.wikipedia.org/wiki/Clique_(graph_theory)
        cliques = []
        for p in sorted(compat):
            vert = p
            if compat[p] == set():
                cliques.append(set([p]))
                continue
            for l in list(compat[p]):
                edge = (p,l)
                processed = False
                for e, c in enumerate(cliques):
                    if p in c and (l in c or all(l in compat[y] for y in c)):
                        c.add(l)
                        processed = True
                if not processed:
                    cliques.append(set([p,l]))

        elevations = []
        for c in cliques:
            elevations.append(sum(map(elevation, c)))

        if elevations:
            eleven = max(elevations)
        else :
            eleven = 0

        return eleven

def enumerate_pil(pilstring, 
        is_file = False, 
        enumfile = None,
        detailed = True, 
        condensed = False, 
        enumconc = 'M', 
        **kwargs):
    """ A wrapper function to directly enumerate a pilstring or file.

    Args:
        pilstring (str): Either a full pil string or a file name containing that string.
        is_file (bool, optional): True if pilstring is a filename, False otherwise. 
            Defaults to False.
        enumfile (str, optional): A filename for the enumerated output. The function 
            returns the output string if no enumfile is specified. Defaults to None.
        detailed (bool, optional): Returns the detailed reaction network. Defaults to True.
        condensed (bool, optional): Returns the condensed reaction network. Defaults to False.
        enumconc (str, optional): Chose concentation units for output: 'M', 'mM', 'uM', 'nM', ...
        **kwargs: Attributes of the peppercornenumerator.Enumerator object.

    Returns:
        Enumerator-object, Outputstring
    """
    cxs, rxns, comp = read_pil(pilstring, is_file, composite = True)

    init_cplxs = list(filter(lambda x: x._concentration is None or \
                    float(x._concentration[1]) != (0.0), cxs.values()))

    enum = Enumerator(init_cplxs, rxns)

    # set kwargs parameters
    for k, w in kwargs.items():
        if hasattr(enum, k):
            setattr(enum, k, w)
        else:
            raise PeppercornUsageError('No Enumerator attribute called: {}'.format(k))

    enum.enumerate()

    outstring = enum.to_pil(enumfile, detailed = detailed, condensed = condensed, 
                composite = comp, molarity = enumconc)
    
    return enum, outstring

def enumerate_ssw(sswstring, 
        is_file = False, 
        ssw_expl = False,
        ssw_conc = 100e-9,
        ssw_rxns = 'T25-utbr-leak-reduced',
        dry_run = True,
        enumfile = None,
        detailed = True, 
        condensed = False, 
        enumconc = 'M', 
        **kwargs):
    """ A wrapper function to directly enumerate a seesaw string or file.

    Args:
        sswstring (str): Either a full seesaw string or a file name containing that string.
        is_file (bool, optional): True if sswstring is a filename, False otherwise. 
            Defaults to False.

        ssw_expl (bool, optional): Explict domain-level specification? Defaults to False.
        ssw_conc (str, optional): 1x concentation in 'M'. Defaulst to 1e-7.
        ssw_rxns (str, optional): Specify the kind of seesaw reactions. Defaults to 
            T25-utbr-leak-reduced.
        dry_run (bool, optional): True if peppercorn enumeration is desired. Defaults to False.
        enumfile (str, optional): A filename for the output. The function 
            returns the output string if no enumfile is specified. Defaults to None.
        detailed (bool, optional): Returns the detailed reaction network. Defaults to True.
        condensed (bool, optional): Returns the condensed reaction network. Defaults to False.
        enumconc (str, optional): Chose concentation units for output: 'M', 'mM', 'uM', 'nM', ...
        **kwargs: Attributes of the peppercornenumerator.Enumerator object.

    Returns:
        Enumerator-object, Outputstring
    """
    utbr = kwargs.get('utbr_species', True)
    cxs, rxns = read_seesaw(sswstring, is_file, 
            explicit = ssw_expl,
            conc = ssw_conc,
            reactions = ssw_rxns)

    enum = Enumerator(list(cxs.values()), rxns)

    # set kwargs parameters
    for k, w in kwargs.items():
        if hasattr(enum, k):
            setattr(enum, k, w)
        else:
            raise PeppercornUsageError('No Enumerator attribute called: {}'.format(k))

    if dry_run:
        enum.dry_run()
    else:
        enum.enumerate()

    outstring = enum.to_pil(enumfile, detailed = detailed, condensed = condensed, molarity = enumconc)
   
    return enum, outstring


def segment_neighborhood(complexes, reactions, p_min=None, represent=None):
    """
    Segmentation of a potentially incomplete neighborhood. That means only the
    specified complexes are interesting, all others should not be returned.

    Beware: Complexes must contain all reactants in reactions *and* there
    must not be any incoming fast reactions into complexes, other than
    those specified in reactions. Thus, we can be sure that the SCCs found here
    are consistent with SCCs found in a previous iteration.

    Args:
        complexes(list[:obj:`PepperComplex`])
    """
    index = 0
    S = []
    SCCs = []

    total = complexes[:]
    for rxn in reactions: 
        total += rxn.products

    total = list(set(total))

    # Set up for Tarjan's algorithm
    for c in total: c._index = None

    # filters reaction products such that there are only species from within complexes
    # this is ok, because it must not change the assignment of SCCs.
    rxns_within = {k: [] for k in complexes}
    rxns_consuming = {k: [r for r in reactions if (k in r.reactants)] for k in total}
    for rxn in reactions:
        assert len(rxn.reactants) == 1
        for product in rxn.products:
            if product in complexes:
                rxns_within[rxn.reactants[0]].append(product)
        rxns_consuming[rxn.reactants[0]]

    def tarjans_scc(cplx, index):
        """
        Executes an iteration of Tarjan's algorithm (a modified DFS) starting
        at the given node.
        """
        # Set this node's tarjan numbers
        cplx._index = index
        cplx._lowlink = index
        index += 1
        S.append(cplx)

        for product in rxns_within[cplx]:
            # Product hasn't been traversed; recurse
            if product._index is None :
                index = tarjans_scc(product, index)
                cplx._lowlink = min(cplx._lowlink, product._lowlink)

            # Product is in the current neighborhood
            elif product in S:
                cplx._lowlink = min(cplx._lowlink, product._index)
    
        if cplx._lowlink == cplx._index:
            scc = []
            while True:
                next = S.pop()
                scc.append(next)
                if next == cplx:
                    break
            SCCs.append(scc)
        return index

    # We now perform Tarjan's algorithm, marking nodes as appropriate
    for cplx in complexes:
        if cplx._index is None:
            tarjans_scc(cplx, index)

    resting_macrostates = []
    transient_macrostates = []
    resting_complexes = []
    transient_complexes = []

    for scc in SCCs:
        try:
            if represent :
                rcx = [cx for cx in scc if cx in represent]

            rcx = sorted(rcx)[0] if rcx else None
            ms = PepperMacrostate(scc[:], prefix='', representative = rcx)
        except DSDDuplicationError as e:
            assert set(e.existing.complexes) == set(scc)
            ms = e.existing

        for c in scc:
            for rxn in rxns_consuming[c]:
                ms.add_reaction(rxn)

        if ms.is_transient:
            transient_complexes += (scc)
        else :
            resting_macrostates.append(ms)

            if p_min:
                for (c, s) in ms.get_stationary_distribution():
                    if s < p_min:
                        transient_complexes.append(c)
                    else :
                        resting_complexes.append(c)
            else:
                resting_complexes += (scc)

    resting_macrostates.sort()
    resting_complexes.sort()
    transient_complexes.sort()

    return {
        'resting_macrostates': resting_macrostates,
        'resting_complexes': resting_complexes,
        'transient_complexes': transient_complexes
    }

def local_elevation_rate(rxn, el=None):
    """Re-calculate a transition rate based on local elevation.

    Right now, downhill reactions are only passed through, but that
    probably needs some more thought about the consequences.

    """
    if el is None:
        assert rxn.arity[0] == 1
        cplx = rxn.reactants[0]
        if cplx._elevation is None:
            return None
        el = cplx._elevation

    if rxn.arity != (1,1) or rxn.rtype in ('branch-3way', 'branch-4way'): 
        # we don't know the reverse rate, ...
        return 1/(1+math.e**(el)) * rxn.const

    elif rxn.const < rxn.reverse_reaction.const:
        # this is a true uphill reaction...
        return 1/(1+math.e**(el)) * rxn.const
    
    # it is a downhill reaction, elevation does not matter ...
    assert rxn.arity[0] == 1
    return rxn.const
 



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name='peppercornenumerator',
    version='0.9',
    description='Domain-level nucleic acid reaction enumerator',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Stefan Badelt, Casey Grun, Karthik Sarma, Brian Wolfe, Seung Woo Shin and Erik Winfree',
    author_email='winfree@caltech.edu',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        ],
    python_requires='>=2.7',
    test_suite='tests',
    install_requires=[
        'future',
        'pandas', # for case-studies
        'numpy',
        'crnsimulator>=0.6',
        'dsdobjects>=0.7.1'],
    packages=['peppercornenumerator'],
    entry_points = {
        'console_scripts': [
            'peppercorn=peppercornenumerator.peppercorn:main',
            'pilsimulator=peppercornenumerator.pilsimulator:main'
            ],
        }
)




#
#  peppercornenumerator/utils.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division

import logging
log = logging.getLogger(__name__)

import re
import sys

class PeppercornUsageError(Exception):
    """Error class to catch usage errors."""

    def __init__(self, msg, val=None):
        self.message = msg
        if val :
            self.message += " ({})".format(val)
        super(PeppercornUsageError, self).__init__(self.message) 

def wrap(x, m):
    """
    Mathematical modulo; wraps x so that 0 <= wrap(x,m) < m. x can be negative.
    """
    return (x % m + m) % m

def natural_sort(l):
    """
    Sorts a collection in the order humans would expect. Implementation from
    http://stackoverflow.com/questions/4836710/does-python-have-a-built-in-function-for-string-natural-sort
    """
    def convert(text): 
        return int(text) if text.isdigit() else text.lower()

    def alphanum_key(key): 
        return [convert(c) for c in re.split('([0-9]+)', str(key))]

    return sorted(l, key=alphanum_key)

def find(f, seq, default=None):
    """
    Return first item in sequence where f(item) == True.
    """
    for item in seq:
        if f(item):
            return item
    return default

def index_parts(enum):
    """
    Testing tool. Accepts an enumerator, produces a tuple
    (domains, strands, complexes) where each element is a dict mapping names
    of those objects to the objects in the enumerator. For instance, domains
    maps the name of each domain in the enumerator to the Domain object
    """
    domains = {}
    strands = {}
    complexes = {}

    for domain in enum.domains:
        domains[domain.name] = domain

    for strand in enum.strands:
        strands[strand.name] = strand

    for complex in enum.initial_complexes:
        complexes[complex.name] = complex

    return (domains, strands, complexes)

class Loop(object):
    """
    Represents (possibly a part) of a single open or closed loop

    Args: 
        loop [(Domain, bound_to, position),..]: Is a list of descriptors for
            involved domains.  In particular: [0] A Domain Object, [1] its
            binding partner position (or None), [2] the domain position.
            None marks the 3' or 5' end of a strand.
    
    TODO: Loop == Loop is only true if the two loops were initialized with the
    same domain ordering.

    """

    def __init__(self, loop):
        self._parts = loop

        is_open = False
        bases = 0
        stems = 0

        # Sanity check to alert that a stem has been specified with both
        # complementary domains.
        stem_list = set()

        # calculate stems, bases, and is_open   #loop re-written by EW
        for step in loop:
            if step is None:
                is_open = True
            else:
                (dom, struct, loc) = step
                if struct is None:
                    bases += len(dom)
                elif struct in stem_list:
                    raise PeppercornUsageError('Double stem count in Loop() Object.')
                else :
                    stems += 1
                    stem_list.add(loc)

        # update cached properties
        self._is_open = is_open
        self._bases = bases
        self._stems = stems

    @property
    def locs(self):
        return (part[2] if part is not None else None for part in self._parts)

    @property
    def domains(self):
        return (part[0] if part is not None else None for part in self._parts)

    @property
    def structures(self):
        return (part[1] if part is not None else None for part in self._parts)

    @property
    def structs(self):
        raise DeprecationWarning('Use of Loop.structs has been replaced by Loop.structures.')
        return self.structures

    @property
    def parts(self):
        """
        Gives the list of (dom, struct, loc) tuples associated with this loop
        """
        return self._parts[:]

    @property
    def stems(self):
        """
        Gives the number of stems in the loop
        """
        return self._stems

    @property
    def bases(self):
        """
        Gives the number of bases in the loop
        """
        return self._bases

    @property
    def is_open(self):
        """
        True if the loop is an open loop, else false
        """
        return self._is_open

    def __contains__(self, item):
        if isinstance(item, Domain):
            return item in self.domains
        elif isinstance(item, tuple):
            return item in self.locs

    def __str__(self):
        return ' '.join([(str(s) if s is not None else '+') for s in self.domains])

    def __repr__(self):
        return "Loop({:s})".format(list(self.domains))

    def __len__(self):
        return sum(len(dom) for dom in self.domains)

    def __eq__(self, other):
        return self._parts == other._parts




#
#  peppercornenumerator/objects.py
#  EnumeratorProject
#
from __future__ import absolute_import, print_function, division
from builtins import map

import logging
log = logging.getLogger(__name__)

import numpy as np
from collections import namedtuple

from dsdobjects import clear_memory
from dsdobjects import DSDObjectsError, DSDDuplicationError
from dsdobjects.core import DL_Domain, DSD_Complex, DSD_Reaction, DSD_Macrostate
from dsdobjects.utils import split_complex, convert_units
from dsdobjects.utils import make_pair_table, pair_table_to_dot_bracket 

class PepperObjectsError(Exception):
    """
    pepperobjects error class.
    """

    def __init__(self, message, *kargs):
        if kargs:
            self.message = "{} [{}]".format(message, ', '.join(map(str,kargs)))
        else :
            self.message = message
        super(PepperObjectsError, self).__init__(self.message)

class PepperDomain(DL_Domain):
    """
    Represents a single domain. We allow several options for specifying domain
    properties. Domains might have an explicit integer (bp) length, or may be
    designated as short or long. If the latter method is used, the code will use
    the relevant constant as the integer domain length.
    """

    def __new__(cls, name, dtype=None, length=None):
        # The new method returns the present instance of an object, if it exists
        self = DL_Domain.__new__(cls)
        try:
            super(PepperDomain, self).__init__(name, dtype, length)
        except DSDDuplicationError as e :
            other = e.existing
            if dtype and (other.dtype != dtype) :
                raise PepperObjectsError('Conflicting dtype assignments for {}: "{}" vs. "{}"'.format(
                    name, dtype, other.dtype))
            elif length and (other.length != length) :
                raise PepperObjectsError('Conflicting length assignments for {}: "{}" vs. "{}"'.format(
                    name, length, other.length))
            return e.existing
        self._nucleotides = None
        return self

    def __init__(self, name, dtype=None, length=None):
        # Remove default initialziation to get __new__ to work
        pass

    @property
    def nucleotides(self):
        return self._nucleotides

    @nucleotides.setter
    def nucleotides(self, value):
        self._nucleotides = value

    @property
    def complement(self):
        # If we initialize the complement, we need to know the class.
        if self._complement is None:
            cname = self._name[:-1] if self.is_complement else self._name + '*'
            if cname in DL_Domain.MEMORY:
                self._complement = DL_Domain.MEMORY[cname]
            else :
                self._complement = PepperDomain(cname, self.dtype, self.length)
        return self._complement

    def can_pair(self, other):
        """
        Returns True if this domain is complementary to the argument.
        """
        return self == ~other

    @property
    def identity(self):
        """
        Returns the identity of this domain, which is its name without a
        complement specifier (i.e. A and A* both have identity A).
        """
        return self._name[:-1] if self._name[-1] == '*' else self._name

    @property
    def is_complement(self):
        """
        Returns true if this domain is a complement (e.g. A* rather than A),
        false otherwise.
        """
        return self._name[-1:] == '*'

class PepperComplex(DSD_Complex):
    """
    Peppercorn complex object. 

    Overwrites some functions with new names, adds some convenient stuff..
    """

    PREFIX = 'e'
    CONCENTRATION = namedtuple('concentration', 'mode value unit')

    @staticmethod
    def clear_memory(memory=True, names=True, ids=True):
        if memory:
            DSD_Complex.MEMORY = dict()
        if names:
            DSD_Complex.NAMES = dict()
        if ids:
            DSD_Complex.ID = dict()

    def __init__(self, sequence, structure, name='', prefix='', memorycheck=True):
        try :
            if not prefix :
                prefix = PepperComplex.PREFIX
            super(PepperComplex, self).__init__(sequence, structure, name, prefix, memorycheck)
        except DSDObjectsError :
            backup = 'enum' if prefix != 'enum' else 'pepper'
            super(PepperComplex, self).__init__(sequence, structure, name, backup, memorycheck)
            log.warning('Complex name existed, prefix has been changed to: {}'.format(backup))
        
        # Peppercorn IO:
        self._concentration = None
        self._elevation = None
        assert self.is_domainlevel_complement

    @property
    def concentration(self):
        if self._concentration is not None:
            return self._concentration
        return None

    @concentration.setter
    def concentration(self, trip):
        if trip is None:
            self._concentration = None
        else:
            (mode, value, unit) = trip
            assert isinstance(value, (int, float))
            self._concentration = PepperComplex.CONCENTRATION(mode, value, unit)

    def concentrationformat(self, out):
        mod = self._concentration.mode
        val = self._concentration.value
        uni = self._concentration.unit
        val = convert_units(val, uni, out) 
        return PepperComplex.CONCENTRATION(mod, val, out)

    @property
    def pair_table(self):
        return super(PepperComplex, self).pair_table
    
    @pair_table.setter
    def pair_table(self, pt):
        self._pair_table = pt

    def full_string(self):
        return "Complex(%s): %s %s" % (
            self.name, str(self.strands), str(self.structure))

    def get_structure(self, loc):
        return self.get_paired_loc(loc)

    def triple(self, *loc):
        # overwrite standard func
        return (self.get_domain(loc), self.get_paired_loc(loc), loc)

    @property
    def available_domains(self):
        ad = []
        for (x,y) in self.exterior_domains:
            ad.append((self.get_domain((x,y)), x, y))
        return ad

    @property
    def pk_domains(self):
        pd = []
        for (x,y) in self.enclosed_domains:
            pd.append((self.get_domain((x,y)), x, y))
        return pd


    def rotate_location(self, loc, n=None):
        return self.rotate_pairtable_loc(loc, n)

    def split(self):
        """ Split PepperComplex into disconneted components.
        """
        if self.is_connected:
            return [self]
        else :
            ps = self.lol_sequence
            pt = self.pair_table
            parts = split_complex(ps, pt)
            cplxs = []
            # assign new_complexes
            for (se,ss) in parts:
                try:
                    cplxs.append(PepperComplex(se, ss))
                except DSDDuplicationError as e:
                    cplxs.append(e.existing)
            return sorted(cplxs)

class PepperReaction(DSD_Reaction):
    RTYPES = set(['condensed', 'open', 'bind11', 'bind21', 'branch-3way', 'branch-4way'])

    def __init__(self, *kargs, **kwargs):
    #def __init__(self, reactants, products, rtype=None, rate=None, memorycheck=True):
        super(PepperReaction, self).__init__(*kargs, **kwargs)
        if self._rtype not in PepperReaction.RTYPES:
            try:
                del DSD_Reaction.MEMORY[self.canonical_form]
            except KeyError:
                pass
            raise DSDObjectsError('Reaction type not supported! ' + 
            'Set supported reaction types using PepperReaction.RTYPES')

        self._reverse_reaction = None

        # Store the 4 relevant Loop() objects (reactant based):
        #   * initial-locus
        #   * target-locus
        #   * x-linker
        #   * y-linker
        self.meta = None
        self.rotations = None

    @property
    def reverse_reaction(self):
        return self._reverse_reaction

    @reverse_reaction.setter
    def reverse_reaction(self, rxn):
        def rev_rtype(rtype, arity):
            """Returns the reaction type of a corresponding reverse reaction. """
            if rtype == 'open' and arity == (1,1):
                return 'bind11'
            elif rtype == 'open' and arity == (1,2):
                return 'bind21'
            elif rtype == 'bind11' or rtype == 'bind21':
                return 'open'
            else:
                return rtype

        if rxn is not False:
            assert rxn.rtype == rev_rtype(self.rtype, self.arity)
        self._reverse_reaction = rxn

    def full_string(self, molarity='M', time='s'):
        """Prints the reaction in PIL format.
        Reaction objects *always* specify rate in /M and /s.  """

        if self.rate :
            newunits = [molarity] * (self.arity[0] - 1) + [time]
            newrate = self.rateformat(newunits)
            rate = newrate.constant
            assert newunits == newrate.units
            units = ''.join(map('/{}'.format, newrate.units))
        else :
            rate = float('nan')
            units = ''

        if self.rtype :
            return '[{:14s} = {:12g} {:4s} ] {} -> {}'.format(self.rtype, rate, units,
                    " + ".join(map(str, self.reactants)), " + ".join(map(str, self.products)))
        else :
            return '[{:12g} {:4s} ] {} -> {}'.format(rate, units,
                    " + ".join(map(str, self.reactants)), " + ".join(map(str, self.products)))

class PepperMacrostate(DSD_Macrostate):
    def __init__(self, *kargs, **kwargs):
        super(PepperMacrostate, self).__init__(*kargs, **kwargs)
        self._internal_reactions = set()
        self._exit_reactions = set()

        # all reactions consuming a particular species
        #self._reactions_consuming = dict()
        self._stationary_distribution = []
        self._exit_probabilities = []

    @property
    def is_resting(self):
        """True iff it is a resting macrostate. """
        return len(self._exit_reactions) == 0

    @property
    def is_transient(self):
        """True iff it is a transient macrostate. """
        return not self.is_resting

    def add_reaction(self, rxn):
        """Adds a reaction to the macrostate. """
        assert len(rxn.reactants) == 1
        assert rxn.reactants[0] in self._complexes
        is_internal = rxn.products[0] in self._complexes
        if is_internal:
            assert len(rxn.products) == 1
            self._internal_reactions.add(rxn)
        else :
            self._exit_reactions.add(rxn)

        self._stationary_distribution = []
        self._exit_probabilities = []

    def get_exit_probabilities(self, warnings=True):
        """
        """
        # build set and list of elements in SCC; assign a numerical index to each complex
        scc_set = frozenset(self._complexes)
        complex_indices = {c: i for (i, c) in enumerate(self._complexes)}
    
        # sort reactions into internal and outgoing; assign a numerical index to each reaction
        r_outgoing = self._exit_reactions
        r_internal = self._internal_reactions
        exit_indices = {r: i for (i, r) in enumerate(r_outgoing)}
    
        # L = # of complexes in SCC
        L = len(self._complexes)
    
        # e = # of exit pathways
        e = len(r_outgoing)
    
        # add transition rates for each internal reaction
        T = np.zeros((L, L))
        for r in r_internal:
            assert len(r.reactants) == 1
            assert len(r.products) == 1
            a = r.reactants[0]
            b = r.products[0]
            T[complex_indices[a]][complex_indices[b]] = r.rate

        # add transition rates for each outgoing reaction
        Te = np.zeros((L, e))
        for r in r_outgoing:
            a = r.reactants[0]
            Te[complex_indices[a]][exit_indices[r]] = r.rate
    
        # the full transition matrix P_{L+e x L+e} would be
        #
        # P = ( Q_{LxL}  R_{Lxe}   )
        #     ( 0_{exL}  I_{exe}   )
        #
        # but we really only care about Q to calculate the fundamental matrix,
        # so we construct
        #
        # P = (T_{LxL} Te_{Lxe})
        P = np.hstack((T, Te))
    
        # then normalize P along each row, to get the overall transition
        # probabilities, e.g. P_ij = P(i -> j), where i,j in 0...L+e
        P = P / np.sum(P, 1)[:, np.newaxis]
    
        # extract the interior transition probabilities (Q_{LxL})
        Q = P[:, 0:L]
    
        # extract the exit probabilities (R_{Lxe})
        R = P[:, L:L + e]
    
        # calculate the fundamental matrix (N = (I_L - Q)^-1)
        N = np.linalg.inv(np.eye(L) - Q)

        # make sure all elements of fundamental matrix are >= 0
        if not (N >= 0).all() :  # --- commented out by EW (temporarily)
            log.error('Negative elements in fundamental matrix. Condensed reaction rates may be incorrect.')
    
        # calculate the absorption matrix (B = NR)
        B = np.dot(N, R)

        # --- added by EW as a weaker surrugate for the above, when necessary
        # assert (B >= 0).all()
    
        # return dict mapping tuples of (incoming complex, outgoing reaction)
        # to exit probabilities
        return {(c, r): B[i, j] for (c, i) in complex_indices.items()
                for (r, j) in exit_indices.items()}


    def get_stationary_distribution(self, warnings=True):
        """
        Take a strongly connected component and calculate the stationary distribution.

        Args:
            T (numpy.matrix): a rate matrix 
            nodes (list, optional): A list of objects for which stationary distribution is to
                be determined.

        Returns:
            [:obj:`dict()`]: Stationary distributions: dict['cplx'] = sdist (flt)
        """
        if not self._stationary_distribution :

            reactions = self._internal_reactions

            # Initialize a Transition Matrix where numerical index corresponds to each complex
            L = len(self._complexes)
            indices = {c: i for (i, c) in enumerate(self._complexes)}
            T = np.zeros((L, L))

            for rxn in reactions:
                # r : a -> b
                # T_{b,a} = rate(r : a -> b)
                #if rxn.reactants[0] in indices and rxn.products[0] in indices:
                a = rxn.reactants[0]
                b = rxn.products[0]
                T[indices[b]][indices[a]] = rxn.rate

            # compute diagonal elements of T
            T_diag = np.sum(T, axis=0)  # sum over columns
            for i in range(L):
                T[i][i] = -T_diag[i]

            # calculate eigenvalues
            (w, v) = np.linalg.eig(T)
            # w is array of eigenvalues
            # v is array of eigenvectors, where column v[:,i] is eigenvector
            # corresponding to the eigenvalue w[i].

            # find eigenvector corresponding to eigenvalue zero (or nearly 0)
            epsilon = 1e-5
            i = np.argmin(np.abs(w))
            if abs(w[i]) > epsilon:
                log.warn(
                    ("Bad stationary distribution for resting set transition matrix. " +
                     "Eigenvalue found %f has magnitude greater than epsilon = %f. " +
                     "Markov chain may be periodic, or epsilon may be too high. Eigenvalues: %s") %
                    (w(i), epsilon, str(w)))
            s = v[:, i]

            # check that the stationary distribution is good
            if warnings and not ((s >= 0).all() or (s <= 0).all()) : 
                #for cl,sd in zip(self._complexes, s):
                #    print(cl, '=', cl.kernel_string, sd)
                #for rxn in reactions:
                #    print(rxn, rxn.rate)
                log.error('Stationary distribution of resting set complex' +
                        'should not be an eigenvector of mixed sign.')

            self._stationary_distribution = s / np.sum(s)

            if not (abs(np.sum(self._stationary_distribution) - 1) < epsilon) :
                log.error('Stationary distribution of resting set complex' +
                        'should sum to 1 after normalization. Condensed reaction' +
                        'rates may be incorrect.')

        return list(zip(self._complexes, self._stationary_distribution))

